-----------------------
-- RaidKiller        --
-- By: Wire @ Byriel --
-----------------------

RaidKiller.State.Encounter = class(RaidKiller.Category.Object);

-- constants
RaidKiller.State.Encounter.Constants = {
  Persistence = {
    History = "History"
  }
};

-- defaults
RaidKiller.State.Encounter.Defaults = {
  Timeout = 10
};

function RaidKiller.State.Encounter:Constructor(data)
  RaidKiller.Category.Object.Constructor(self, data);
  
  -- parameters
  self:SetPhases(data.phases);
  
  -- unit states
  local unitStates = data.units;
  if (not unitStates) then
    unitStates = {
      RaidKiller.State.Unit({
        primary = true,
        unit = data.name
      })
    };
  end
  self:SetUnitStates(unitStates);
  
  -- timeout
  local timeout = RaidKiller.State.Encounter.Defaults.Timeout;
  if (data.timeout) then
    timeout = data.timeout;
  end
  self:SetTimeout(timeout);
end

-- manager
function RaidKiller.State.Encounter:GetManager()
  return self.manager;
end

function RaidKiller.State.Encounter:SetManager(value)
  self.manager = value;
  
  -- update unit states
  for k, state in pairs(self:GetUnitStates()) do
    state:SetManager(self:GetManager());
  end
  
  -- update phases
  for k, phase in pairs(self:GetPhasesByOrder()) do
    phase:SetManager(self:GetManager());
  end
end

-- encounter
function RaidKiller.State.Encounter:GetEncounter()
  return self.encounter;
end

function RaidKiller.State.Encounter:SetEncounter(value)
  self.encounter = value;
end

-- timeout
function RaidKiller.State.Encounter:GetTimeout()
  return self.timeout;
end

function RaidKiller.State.Encounter:SetTimeout(value)
  self.timeout = value;
end

-- engagement time
function RaidKiller.State.Encounter:GetEngagementTime()
  return self.engagementTime;
end

function RaidKiller.State.Encounter:SetEngagementTime(value)
  self.engagementTime = value;
end

-- unit states
function RaidKiller.State.Encounter:GetUnitStates()
  if (type(self.unitStates) ~= "table") then
    self.unitStates = {};
  end
  return self.unitStates;
end

function RaidKiller.State.Encounter:SetUnitStates(value)
  self.unitStates = value;
  
  -- register
  for k, state in pairs(self:GetUnitStates()) do
    state:SetState(self);
  end
end

-- phases
function RaidKiller.State.Encounter:GetPhasesByName()
  if (type(self.phasesByName) ~= "table") then
    self.phasesByName = {};
  end
  return self.phasesByName;
end

function RaidKiller.State.Encounter:GetPhasesByOrder()
  if (type(self.phasesByOrder) ~= "table") then
    self.phasesByOrder = {};
  end
  return self.phasesByOrder;
end

function RaidKiller.State.Encounter:SetPhases(value)
  
  -- phases
  for k, phase in pairs(value) do
    
    -- index by name
    self:GetPhasesByName()[phase:GetName()] = phase;
    
    -- index by order
    self:GetPhasesByOrder()[phase:GetOrder()] = phase;
    
    -- register
    phase:SetState(self);
  end
end

-- duration
function RaidKiller.State.Encounter:GetDuration()
  local result = 0;
  if (self:GetEngagementTime()) then
    result = math.floor(Inspect.Time.Frame() - self:GetEngagementTime());
  end
  return result;
end

-- phase
function RaidKiller.State.Encounter:GetPhase()
  return self.phase;
end

function RaidKiller.State.Encounter:SetPhase(value)
  
  -- changed
  local changed = false;
  if (value ~= self:GetPhase()) then
    changed = true;
  end
  
  -- update manager
  if (changed and self:GetPhase() and value) then
    self:GetManager():PhaseExited(self:GetPhase());
  end
  
  -- save value
  self.phase = value;
  
  -- update manager
  if (changed and self:GetPhase()) then
    self:GetManager():PhaseEntered(self:GetPhase());
  end
end

-- phase
function RaidKiller.State.Encounter:PhaseCompleted(phase)
  
  -- next phase
  self:SetPhase(self:GetPhasesByOrder()[phase:GetOrder() + 1]);
  
  -- complete
  if (not self:GetPhase()) then
    self:Completed();
  end
end

-- engaged
function RaidKiller.State.Encounter:Engaged()
  
  -- time
  local time = Inspect.Time.Frame();
  
  -- engagement time
  self:SetEngagementTime(time);
  
  -- attempts
  self:GetHistory().Attempts.Total = self:GetHistory().Attempts.Total + 1;
  
  -- clear phase
  self:SetPhase();
  
  -- reset phases
  for k, phase in pairs(self:GetPhasesByOrder()) do
    phase:Reset();
  end
  
  -- phase one
  self:SetPhase(self:GetPhasesByOrder()[1]);
  
  -- debug
  self:Debug("Encounter: " .. self:GetEncounter():GetName() .. " started.");
  
  -- update manager
  self:GetManager():Engaged();
end

-- failed
function RaidKiller.State.Encounter:Failed()
  
  -- attempts
  local attempts = self:GetHistory().Attempts;
  
  -- last
  attempts.Last = math.ceil(Inspect.Time.Frame() - self:GetTimeout() - self:GetEngagementTime());
  
  -- first
  if (attempts.Best == 0) then
    attempts.Best = attempts.Last;
  end
  
  -- best
  if (attempts.Last > attempts.Best) then
    attempts.Best = attempts.Last;
  end
  
  -- debug
  self:Debug("Encounter: " .. self:GetEncounter():GetName() .. " failed! Attempted " .. attempts.Total .. " time(s).");
  self:Debug("Best Time: " .. attempts.Best);
  self:Debug("Last Time: " .. attempts.Last);
  
  -- stop encounter
  self:GetManager():StopEncounter();
end

function RaidKiller.State.Encounter:Completed()
  
  -- completions
  local completions = self:GetHistory().Completions;
  
  -- total
  completions.Total = completions.Total + 1;
  
  -- last
  completions.Last = math.ceil(Inspect.Time.Frame() - self:GetEngagementTime());
  
  -- last
  self:GetHistory().Attempts.Last = completions.Last;
  
  -- first
  if (completions.Best == 0) then
    completions.Best = completions.Last;
  end
  
  -- best
  if (completions.Last < completions.Best) then
    completions.Best = completions.Last;
  end
  
  -- debug
  self:Debug("Encounter: " .. self:GetEncounter():GetName() .. " completed " .. completions.Total .. " time(s)!");
  self:Debug("Best Time: " .. completions.Best);
  self:Debug("Last Time: " .. completions.Last);
  
  -- update manager
  self:GetManager():StopEncounter();
end

-- persistence

-- history
function RaidKiller.State.Encounter:GetHistory()
  if (type(self:GetPersistenceValue(RaidKiller.State.Encounter.Constants.Persistence.History)) ~= "table") then
    self:SetPersistenceValue(RaidKiller.State.Encounter.Constants.Persistence.History, {
      Attempts = {
        Total = 0,
        Best = 0,
        Last = 0
      },
      Completions = {
        Total = 0,
        Best = 0,
        Last = 0
      }
    });
  end
  return self:GetPersistenceValue(RaidKiller.State.Encounter.Constants.Persistence.History);
end

-- debug
function RaidKiller.State.Encounter:Debug(value)
  self:GetManager():GetApplication():Debug(value);
end
